"""
   Copyright 2021 Ian Housman

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

#Example of how to run CCDC and view outputs using the Python visualization tools
#Acquires Landsat data, runs CCDC, and tries to add them to the viewer
#Original CCDC paper: https://www.sciencedirect.com/science/article/pii/S0034425714000248
#Since CCDC doesn't work well on-the-fly, see the CCCDCViz.py example to view outputs created with this script
#The general workflow for CCDC is to run this script, and then either utilize the harmonic model for a given date
#or to use the breaks for change detection. All of this is demonstrated in the CCDCViz.py example
####################################################################################################
import os,sys
sys.path.append(os.getcwd())

#Module imports
import geeViz.getImagesLib as getImagesLib
ee = getImagesLib.ee
Map = getImagesLib.Map
Map.clearMap()
####################################################################################################
#Define user parameters:
# Define user parameters:

# Specify study area: Study area
# Can be a featureCollection, feature, or geometry
studyArea = getImagesLib.testAreas['CA']

#Update the startJulian and endJulian variables to indicate your seasonal 
#constraints. This supports wrapping for tropics and southern hemisphere.
#If using wrapping and the majority of the days occur in the second year, the system:time_start will default 
#to June 1 of that year.Otherwise, all system:time_starts will default to June 1 of the given year
#startJulian: Starting Julian date 
#endJulian: Ending Julian date
startJulian = 1
endJulian = 365 

#Specify start and end years for all analyses
#More than a 3 year span should be provided for time series methods to work 
#well. If providing pre-computed stats for cloudScore and TDOM, this does not 
#matter
startYear = 1984
endYear = 2021

#Choose whether to include Landat 7
#Generally only included when data are limited
includeSLCOffL7 = True



#Export params
#Whether to export CCDC outputs
exportCCDC = False

#Which bands/indices to export
#These will not always be used to find breaks - that is specified below in the ccdcParams
#Options are: ["blue","green","red","nir","swir1","swir2","NDVI","NBR","NDMI","NDSI","brightness","greenness","wetness","fourth","fifth","sixth","tcAngleBG"]
#Be sure that any bands in ccdcParams.breakpointBands are in this list
exportBands = ["blue","green","red","nir","swir1","swir2","NDVI"]


#Set up Names for the export
outputName = 'CCDC-Test4'

#Provide location composites will be exported to
#This should be an asset folder, or more ideally, an asset imageCollection
exportPathRoot = 'users/iwhousman/test/ChangeCollection'


#CRS- must be provided.  
#Common crs codes: Web mercator is EPSG:4326, USGS Albers is EPSG:5070, 
#WGS84 UTM N hemisphere is EPSG:326+ zone number (zone 12 N would be EPSG:32612) and S hemisphere is EPSG:327+ zone number
crs = 'EPSG:5070'

#Specify transform if scale is None and snapping to known grid is needed
transform = [30,0,-2361915.0,0,-30,3177735.0]

#Specify scale if transform is None
scale = None


###############################################################
#CCDC Params
# ccdcParams ={
#   breakpointBands:['green','red','nir','swir1','swir2','NDVI'],//The name or index of the bands to use for change detection. If unspecified, all bands are used.//Can include: 'blue','green','red','nir','swir1','swir2'
#                                                               //'NBR','NDVI','wetness','greenness','brightness','tcAngleBG'
#   tmaskBands : None,//['green','swir2'],//The name or index of the bands to use for iterative TMask cloud detection. These are typically the green band and the SWIR2 band. If unspecified, TMask is not used. If specified, 'tmaskBands' must be included in 'breakpointBands'., 
#   minObservations: 6,//Factors of minimum number of years to apply new fitting.
#   chiSquareProbability: 0.99,//The chi-square probability threshold for change detection in the range of [0, 1],
#   minNumOfYearsScaler: 1.33,//Factors of minimum number of years to apply new fitting.,\
#   lambda: 0.002,//Lambda for LASSO regression fitting. If set to 0, regular OLS is used instead of LASSO
#   maxIterations : 25000, //Maximum number of runs for LASSO regression convergence. If set to 0, regular OLS is used instead of LASSO.
#   dateFormat : 1 //'fractional' (1) is the easiest to work with. It is the time representation to use during fitting: 0 = jDays, 1 = fractional years, 2 = unix time in milliseconds. The start, end and break times for each temporal segment will be encoded this way.
  
# }; 
ccdcParams ={
  'breakpointBands':['green','red','nir','swir1','swir2','NDVI'],
  'tmaskBands' : None,
  'minObservations': 6,
  'chiSquareProbability': 0.99,
  'minNumOfYearsScaler': 1.33,
  'lambda': 0.002,
  'maxIterations' : 25000,
  'dateFormat' : 1 
}; 

###############################################################
#End user parameters
###############################################################
###############################################################
###############################################################
#Start function calls
###############################################################
#Call on master wrapper function to get Landat and Sentinel 2 scenes
processedScenes = getImagesLib.getProcessedLandsatScenes(studyArea = studyArea,startYear = startYear, endYear = endYear,
                                                        startJulian = startJulian,endJulian = endJulian,
                                                        includeSLCOffL7 = includeSLCOffL7).select(exportBands)


#Remove any extremely high band/index values
def removeGT1(img):
  lte1 = img.select(['blue','green','nir','swir1','swir2']).lte(1).reduce(ee.Reducer.min());
  return img.updateMask(lte1);
processedScenes = processedScenes.map(removeGT1)
Map.addLayer(processedScenes,{},'Processed Input Data',False);

#Set the scene collection in the ccdcParams
ccdcParams['collection'] = processedScenes

#Run CCDC
ccdc = ee.Image(ee.Algorithms.TemporalSegmentation.Ccdc(**ccdcParams))

#Set properties for asset
# ccdc = ccdc.copyProperties(processedScenes)
# ccdc = ccdc.setMulti(ccdcParams)
# ccdc = ee.Image(ccdc)

Map.addLayer(ccdc,{},'CCDC Output',False);

#Export output
if exportCCDC:
  print('Exporting CCDC output')

  exportName = outputName  + '_'+str(startYear) + '_' + str(endYear) + '_' + str(startJulian) + '_' + str(endJulian)
  exportPath = exportPathRoot + '/'+ exportName

  #Export output
  getImagesLib.exportToAssetWrapper(ccdc,exportName,exportPath,{'.default':'sample'},studyArea,scale,crs,transform)

####################################################################################################
#Load the study region
Map.addLayer(studyArea, {'strokeColor': '0000FF'}, "Study Area", False)
Map.centerObject(studyArea)
####################################################################################################
Map.view()


"""
   Copyright 2021 Ian Housman

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
#Script to help visualize variability of observations across space and time
#Intended to work within the geeViz package
######################################################################
from geeViz.getImagesLib import *
import os,json, pdb,glob,math,threading,time,datetime
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patheffects as pe
from datetime import date
###############################################################
def check_dir(dir):
  if not os.path.exists(dir):os.makedirs(dir)
###############################################################
def limitThreads(limit):
  while threading.activeCount() > limit:
    time.sleep(1)
    print(threading.activeCount(),'threads running')
###############################################################
#Function to extract a json table of zonal stats from an image 
def getTableWrapper(image,fc, outputName,reducer = ee.Reducer.first(),scale = 30,crs = 'EPSG:4326',transform = None,tryNumber = 1,maxTries = 15):
  #get zonal stats and export
  table = image.reduceRegions(\
      fc,\
      reducer, \
      scale, \
      crs,\
      transform, \
      4)
  try:
    print('Exporting table:',outputName)
    t = table.getInfo()
    o = open(outputName,'w')
    o.write(json.dumps(t))
    o.close()
    convert_to_csv(outputName)
  except Exception as e:
    print('Error encountered:',e)
   
    tryNumber += 1
    if tryNumber < maxTries and e.args[0].find(" Parameter 'image' is required") == -1:
      print('Trying to convert table again. Try number:',tryNumber)
      getTableWrapper(image,fc, outputName,reducer,scale,crs,transform,tryNumber,maxTries)
###############################################################
#Wrapper to get a sample of locations for a given area
def getTimeSeriesSample(startYear,endYear,startJulian,endJulian,compositePeriod,exportBands,studyArea,nSamples,output_table_name,showGEEViz,maskSnow = False,programs = ['Landsat','Sentinel2']):
  check_dir(os.path.dirname(output_table_name))
  #If available, bring in preComputed cloudScore offsets and TDOM stats
  #Set to null if computing on-the-fly is wanted
  #These have been pre-computed for all CONUS for Landsat and Setinel 2 (separately)
  #and are appropriate to use for any time period within the growing season
  #The cloudScore offset is generally some lower percentile of cloudScores on a pixel-wise basis
  preComputedCloudScoreOffset = getPrecomputedCloudScoreOffsets(10)
  preComputedLandsatCloudScoreOffset = preComputedCloudScoreOffset['landsat']
  preComputedSentinel2CloudScoreOffset = preComputedCloudScoreOffset['sentinel2']

  #The TDOM stats are the mean and standard deviations of the two bands used in TDOM
  #By default, TDOM uses the nir and swir1 bands
  preComputedTDOMStats = getPrecomputedTDOMStats()
  preComputedLandsatTDOMIRMean = preComputedTDOMStats['landsat']['mean']
  preComputedLandsatTDOMIRStdDev = preComputedTDOMStats['landsat']['stdDev']

  preComputedSentinel2TDOMIRMean = preComputedTDOMStats['sentinel2']['mean']
  preComputedSentinel2TDOMIRStdDev = preComputedTDOMStats['sentinel2']['stdDev']

  #####################################################################################
  #Function Calls
  #Get all images
  try:
    saBounds = studyArea.geometry().bounds()
  except:
    saBounds = studyArea.bounds()

  #Sample the study area
  randomSample = ee.FeatureCollection.randomPoints(studyArea, nSamples, 0, 50)
  Map.addLayer(randomSample,{"layerType": "geeVector"},'Samples',True)

  dummyImage = None

  for yr in range(startYear,endYear+1):
    output_table_nameT = '{}_{}_{}_{}-{}_{}_{}{}'.format(os.path.splitext(output_table_name)[0], '-'.join(programs),yr,startJulian,endJulian,compositePeriod,nSamples,os.path.splitext(output_table_name)[1])
    if not os.path.exists(output_table_nameT):
      if 'Landsat' in programs and 'Sentinel2' in programs:
        if dummyImage == None:
          dummyImage = ee.Image(getProcessedLandsatAndSentinel2Scenes(saBounds,\
              2019,\
              2020,\
              1,\
              365).first())

        images = getProcessedLandsatAndSentinel2Scenes(saBounds,\
              yr,\
              yr,\
              startJulian,\
              endJulian,\
              toaOrSR = 'TOA',
              includeSLCOffL7 = True,
              )
      elif 'Sentinel2' in programs: 
        if dummyImage == None:
          dummyImage = ee.Image(getProcessedSentinel2Scenes(saBounds,\
                2019,\
                2020,\
                1,\
                365).first())
        images = getProcessedSentinel2Scenes(saBounds,\
              yr,\
              yr,\
              startJulian,\
              endJulian)
      elif 'Landsat' in programs:
        if dummyImage == None:
          dummyImage = ee.Image(getProcessedLandsatScenes(saBounds,\
                2019,\
                2020,\
                1,\
                365).first())
        images =  getProcessedLandsatScenes(
                  saBounds,\
                  yr,\
                  yr,\
                  startJulian,\
                  endJulian,\
                  toaOrSR = 'TOA',
                  includeSLCOffL7 = True)
      images = fillEmptyCollections(images,dummyImage)

      #Vizualize the median of the images
      # Map.addLayer(images.median(),vizParamsFalse,'Median Composite '+str(yr),False)


      #Mask snow
      if maskSnow:
        print('Masking snow')
        images = images.map(sentinel2SnowMask)


      #Add greenness ratio/indices
      images = images.map(HoCalcAlgorithm2)
      # Map.addLayer(images.select(exportBands),{},'Raw Time Series ' + str(yr),True)

      #Convert to n day composites
      composites = nDayComposites(images,yr,yr,1,365,compositePeriod)

      # Map.addLayer(composites.select(exportBands),{},str(compositePeriod) +' day composites '+str(yr))

      #Convert to a stack
      stack = composites.select(exportBands).toBands();

      #Fix band names to be yyyy_mm_dd
      bns = stack.bandNames()
      bns = bns.map(lambda bn: ee.String(bn).split('_').slice(1,None).join('_'))
      stack = stack.rename(bns);
      
      #Start export table thread
      tt = threading.Thread(target = getTableWrapper, args = (stack,randomSample,output_table_nameT))
      tt.start()
      time.sleep(0.1)
  
  threadLimit = 1
  if showGEEViz:
    #Vizualize outputs
    Map.addLayer(studyArea,{'strokeColor':'00F'},'Study Area')
    Map.centerObject(studyArea)
    Map.view()
    threadLimit = 2
  limitThreads(threadLimit)


  
###############################################################
#Function to convert json gee table into csvs
#Assumes id format is bandName_yyyy-dd-mm
def convert_to_csv(output_table_name):
  with open(output_table_name) as jf:
    table = json.load(jf)
  
  #First find the bands and dates in the table
  bands = []
  dates = []
  print('Finding dates and bands in json:',output_table_name)
  for feature in table['features'][:1]:
    props = feature['properties']
    for prop in list(props.keys()):
      value = props[prop]
      band = prop.split('_')[-1]
      if band not in bands:bands.append(band)

      date = prop.split('_')[0]
      if date not in dates:dates.append(date)
      
  #For each of the bands, create a csv
  for band in bands:
    output_csv = os.path.splitext(output_table_name)[0] + '_{}.csv'.format(band)
    out_table = '{}\n'.format(','.join(dates))
    if not os.path.exists(output_csv):
      print('Parsing:',band)

      #Iterate across each feature and pull all properties and values that have that band
      for feature in table['features']:
        id = feature['id']
        values = []
        props = feature['properties']
        prop_keys = list(props.keys())
        prop_keys = [i for i in prop_keys if i.split('_')[-1] == band]
        
        for prop in prop_keys:
          value = str(props[prop])
          if value == 'None':value = ''
          values.append(value)
        out_line = '{}\n'.format(','.join(values))
        out_table += out_line

      #Write out table
      o = open(output_csv,'w')
      o.write(out_table)
      o.close()

###############################################################
#Function to take a set of csv tables with yyyy-mm-dd dates on the header row and values of a band/index from an
#area for each row (some sort of zonal stat or point location value). Null values are expected to be blank entries in the csv.
#It produces a time series chart of the histogram for each date in the given table
def chartTimeSeriesDistributions(tables,output_dir,output_base_name,n_bins = 40,min_pctl = 1,max_pctl = 99,background_color = '#D6D1CA',font_color = '#1B1716',overwrite = False, howManyHarmonics = 3,showChart = False,annotate_harmonic_peaks = True):
  # print(tables)
  check_dir(output_dir)

  bands = []
  for table in tables:
    band = os.path.splitext(os.path.basename(table))[0].split('_')[-1]
    if band not in bands:bands.append(band)
  for band in bands:
    output_chart_name = os.path.join(output_dir,output_base_name + '_'+band+'.png')
    if not os.path.exists(output_chart_name) or overwrite:
      title = ' '.join(output_base_name.split('_')) + ' '+band + ' Distrubution Time Series'
      tablesT = [table for table in tables if os.path.basename(table).find(band) > -1]
      #Find the name of the band/index
      index_name = band


      print('Creating time series distribution chart for:',band)
      values = []
      data = pd.concat([pd.read_csv(t) for t in tablesT],axis = 1)
      columns = data.columns
      values = data.to_numpy()
     
      #Get min and max
      flat = values.flatten()
      flat = flat[~(np.isnan(flat))]
      min = np.percentile(flat,min_pctl)#flat.min()
      max = np.percentile(flat,max_pctl)#flat.max()
      min_2 = np.percentile(flat,10)
      max_2 = np.percentile(flat,90)
      values = values.clip(min,max)
      #Get dates (yyyy-mm-dd)
      dates = [i.split('_')[0] for i in columns]
      years = np.unique([i.split('-')[0] for i in dates])
      print('years',years)
      # dates = ['2000-{}-{}'.format(i.split('-')[1],i.split('-')[2]) for i in dates]
      
      #Set up bins for histogram
      bin_step = (max-min)/n_bins
      bins = np.arange(min,max+bin_step,bin_step)

      #Get histograms for each date and clip out outlier frequency values
      hist = np.array([np.histogram(data[column], bins=bins,density = True)[0] for column in columns]).transpose()
      hist = np.nan_to_num(hist, nan=0)
      hist = hist.clip(np.percentile(hist,10),np.percentile(hist,99))
      


      #Harmonic regression fitting
      #Get table of all xs and ys
      table_xs = np.array([])
      table_ys = np.array([])
      table_all_xs = np.array([])
      d0 = date(1969, 12, 31)
      percentiles = []
      #Set up tables for harmonic regression
      for i,column in enumerate(columns):
        d = dates[i]
        d1 = date(int(d.split('-')[0]),int(d.split('-')[1]),int(d.split('-')[2]))
        delta = d1 - d0
        delta_fraction = math.modf(delta.days/365.25)[0]
        decimal_date = int(d.split('-')[0])+delta_fraction

        ys = values[:,i]
        ys = ys[~(np.isnan(ys))]
        if(len(ys) > 3):
          percentiles.append(np.percentile(ys,[0,5,25,50,75,95,100]))
        else:
          percentiles.append([np.nan,np.nan,np.nan,np.nan,np.nan,np.nan,np.nan])
        xs = np.repeat(decimal_date,len(ys))
        table_ys = np.append(table_ys,ys)
        table_xs = np.append(table_xs,xs)
        table_all_xs = np.append(table_all_xs,decimal_date)
      percentiles = np.array(percentiles)
      
      table_all_xs = np.array(table_all_xs).flatten()
      table_xs = np.array(table_xs).flatten()
      table_ys = np.array(table_ys).flatten()
 
      #Harmonic regression fit model
      xs = np.array([table_xs]).T
      sin1Term = np.sin(xs*2*math.pi)
      cos1Term = np.cos(xs*2*math.pi)
      sin2Term = np.sin(xs*4*math.pi)
      cos2Term = np.cos(xs*4*math.pi)
      sin3Term = np.sin(xs*6*math.pi)
      cos3Term = np.cos(xs*6*math.pi)
      intTerm = np.ones(xs.shape[0])
      harm_1 = np.c_[sin1Term,cos1Term,xs, intTerm] 
      harm_1_2 = np.c_[sin1Term,cos1Term,sin2Term,cos2Term,xs, intTerm]
      harm_1_2_3 = np.c_[sin1Term,cos1Term,sin2Term,cos2Term,sin3Term,cos3Term,xs, intTerm]

      harm_1_model = np.linalg.lstsq(harm_1, table_ys, rcond=None)
      harm_1_2_model = np.linalg.lstsq(harm_1_2, table_ys, rcond=None)
      harm_1_2_3_model = np.linalg.lstsq(harm_1_2_3, table_ys, rcond=None)
      
      # print(harm_1_model)
      peak_1_fraction = math.atan(harm_1_model[0][0]/harm_1_model[0][1])/(2*math.pi)
      peak_2_fraction = peak_1_fraction + 0.5
      if peak_1_fraction < 0:
        peak_1_fraction = 1+peak_1_fraction
      if peak_2_fraction > 1:
        peak_2_fraction = peak_2_fraction - 1

      peak_1_yr = int(years[-2])+peak_1_fraction
      peak_2_yr = int(years[-2])+peak_2_fraction
      peak_1_pred = np.dot([np.sin(peak_1_yr*2*math.pi),np.cos(peak_1_yr*2*math.pi),peak_1_yr, 1],harm_1_model[0])
      peak_2_pred = np.dot([np.sin(peak_2_yr*2*math.pi),np.cos(peak_2_yr*2*math.pi),peak_2_yr, 1],harm_1_model[0])
      peak_1_y = min_2
      peak_2_y = max_2
      if peak_1_pred > peak_2_pred:
        peak_1_y = max_2
        peak_2_y = min_2

      peak_date = int(peak_1_fraction*365)+1
      peak_date2 = int(peak_2_fraction*365)+1
  

      print(peak_date,peak_date2,peak_1_pred,peak_2_pred,years[-2][2:]+f'{peak_date:03}',years[-2][2:]+f'{peak_date2:03}')
      # print(datetime.datetime.strptime('19'+str(int(peak_date*365)), '%y%j').strftime("%d-%m-%Y"))
      peak_date = datetime.datetime.strptime(years[-2][2:]+f'{peak_date:03}', '%y%j').strftime("%m-%d")
      peak_date2 = datetime.datetime.strptime(years[-2][2:]+f'{peak_date2:03}', '%y%j').strftime("%m-%d")
      # print(peak_date)
      # print(peak_date2)
      #Apply harm model
      xs = np.array([table_all_xs]).T
      sin1Term = np.sin(xs*2*math.pi)
      cos1Term = np.cos(xs*2*math.pi)
      sin2Term = np.sin(xs*4*math.pi)
      cos2Term = np.cos(xs*4*math.pi)
      sin3Term = np.sin(xs*6*math.pi)
      cos3Term = np.cos(xs*6*math.pi)
      intTerm = np.ones(xs.shape[0])
      harm_1 = np.c_[sin1Term,cos1Term,xs, intTerm] 
      harm_1_2 = np.c_[sin1Term,cos1Term,sin2Term,cos2Term,xs, intTerm]
      harm_1_2_3 = np.c_[sin1Term,cos1Term,sin2Term,cos2Term,sin3Term,cos3Term,xs, intTerm]
      # print('beta hat:',beta_hat[0],xs)
      pred_1 = np.dot(harm_1,harm_1_model[0])
      pred_1_2 = np.dot(harm_1_2,harm_1_2_model[0])
      pred_1_2_3 = np.dot(harm_1_2_3,harm_1_2_3_model[0])
      
      pred_dict  = {'1':pred_1,
                    '2':pred_1_2,
                    '3':pred_1_2_3}

      pred = pred_dict[str(howManyHarmonics)]


      #Plot
      xTickFreq = 8#int(len(columns)/20)
      yTickFreq = (max-min)/10#0.1
      width = (len(columns)/25)+2
      if width < 8:
        width = 12
        xTickFreq = 3
      # fig = plt.figure(figsize=(width,6),frameon=True,facecolor='w')
      fig, ax = plt.subplots(figsize=(width,7),frameon=True,facecolor='w')
      fig.patch.set_facecolor(background_color)
      
      params = {"ytick.color" : font_color,
              "xtick.color" : font_color,
              "axes.labelcolor" : font_color,
              "axes.edgecolor" : font_color,
              'legend.fontsize': 7,
              'legend.handlelength': 1.2}
      plt.rcParams.update(params)
      
      # ax = fig.add_axes([0.06, 0.15, 0.84, 0.80])
      ax.set_title(title)
      
      # fig.annotate('local max', xy=(0.5, 0.5))
      cmap = plt.get_cmap('viridis')
      cf = plt.pcolormesh(dates, bins, hist, cmap = cmap)#, vmin = 500)
      degrees = 45
      plt.xticks(rotation=degrees, fontsize = 7,ha = 'right')

      # print(list(zip(dates,pred,table_all_xs)))
      harm_line =plt.plot(dates, pred, linestyle = '-', color = background_color, linewidth = 2, label='Harmonic Fit ({})'.format(howManyHarmonics))
      # median_line = plt.plot(dates, percentiles[:,3], linestyle = '--', color = font_color, linewidth = 1.5,label = 'Median')
  
      ax.set_ylim([min, max])
      ax.set_ylabel('{} Value'.format(index_name), fontsize = 10)
      ax.set_xlabel('Date', fontsize = 10)

      #Set up the x and y axis tick frequencies
      ax.xaxis.set_major_locator(plt.MultipleLocator(xTickFreq))
      ax.xaxis.set_minor_locator(plt.MultipleLocator(1))
      ax.yaxis.set_major_locator(plt.MultipleLocator(yTickFreq))
      ax.yaxis.set_minor_locator(plt.MultipleLocator(1))
      
      ax.grid(True, which='major', axis='y', linestyle='--', color=font_color)
      ax.grid(True, which='major', axis='x', linestyle='--', color=font_color)

      cbax = fig.add_axes([0.93, 0.11, 0.01, 0.71])
      legend = plt.legend(handles = harm_line, bbox_to_anchor=(-2.5, 1.08), loc='upper left')
      
      
      cb = plt.colorbar(cf, cax = cbax, orientation = 'vertical')
      cb.ax.tick_params(labelsize=10)
      cb.set_label('Percent of Samples (%)', rotation = 270, labelpad = 15, fontsize = 10,color=font_color) 

      if peak_1_pred > max:
        peak_1_pred  = max
      elif peak_1_pred < min:
        peak_1_pred  = min
      if peak_2_pred > max:
        peak_2_pred  = max
      elif peak_2_pred < min:
        peak_2_pred  = min
      if(annotate_harmonic_peaks):
        print('Annotating peak dates of harmonics')
        try:
          yr_dates = [i for i in dates if i.split('-')[0] == years[-2]]
          m_dates = [i for i in yr_dates if i.split('-')[1] == peak_date.split('-')[0]]
          if len(m_dates) == 0:
            m_dates = [i for i in yr_dates if int(i.split('-')[1]) == int(peak_date.split('-')[0])-1]
            print(m_dates)
          m_dates = m_dates[0]
          ax.annotate('{} ({})'.format(peak_date,round(peak_1_pred,3)),
                xy=(m_dates, peak_1_pred), xycoords='data',
              # xytext=(m_dates, peak_1_y),    # fraction, fraction
              # textcoords='data',
              # arrowprops=dict(facecolor='black', shrink=0.05),
                color = background_color, fontsize='10',path_effects=[pe.withStroke(linewidth=3, foreground=font_color)]
                )
          yr_dates = [i for i in dates if i.split('-')[0] == years[-2]]
          m_dates = [i for i in yr_dates if i.split('-')[1] == peak_date2.split('-')[0]][0]
          ax.annotate('{} ({})'.format(peak_date2,round(peak_2_pred,3)),
                xy=(m_dates, peak_2_pred), xycoords='data',
                 # xytext=(m_dates, peak_2_y),    # fraction, fraction
              # textcoords='data',
              # arrowprops=dict(facecolor='black', shrink=0.05),
                color = background_color, fontsize='10',path_effects=[pe.withStroke(linewidth=3, foreground=font_color)]
                )
        except Exception as e:
          print(e)
      # ax.annotate('Peak Date 2',
      #       xy=('2019-{}'.format(peak_date2), 0.3), xycoords='data'
      #       )
      # plt.text(0.5, 1.1, 'Test', color=font_color, 
        # bbox=dict(facecolor='none', edgecolor=font_color))
      fig.savefig(output_chart_name)
      if showChart:
        plt.show()
      plt.close()
    else:
      print('Already produced:',output_chart_name)


"""
   Copyright 2020 Ian Housman & Leah Campbell

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
###############################################################################
#           TASKMANAGERLIB.PY
################################################################################
# Functions for GEE task management

import ee, time, re
from datetime import datetime, timedelta
try:
    z = ee.Number(1).getInfo()
except:
    print('Initializing GEE')
    ee.Initialize()


#------------------------------------------------------------------------------
#                     Standard Task Tracking 
#------------------------------------------------------------------------------
# Standard task tracker - prints number of ready and running tasks each 10 seconds
def now():
    return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
def trackTasks():
    x = 1
    while x <1000:
        tasks = ee.data.getTaskList()
        ready = [i for i in tasks if i['state'] == 'READY']
        running = [i for i in tasks if i['state'] == 'RUNNING']
        running_names = [[str(i['description']),str(timedelta(seconds = int(((time.time()*1000)-int(i['start_timestamp_ms']))/1000)))] for i in running]
        now = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        print(len(ready),'tasks ready',now)
        print(len(running),'tasks running',now)
        print('Running names:')
        for rn in running_names:print(rn)
        print()
        print()
        time.sleep(10)
        x+=1
def trackTasks2(credential_name = None,id_list = None,task_count = 1):
  while task_count > 0:
    tasks = ee.data.getTaskList()
    if id_list != None:
      tasks = [i for i in tasks if i['description'] in id_list]
    ready = [i for i in tasks if i['state'] == 'READY']
    running = [i for i in tasks if i['state'] == 'RUNNING']
    running_names = [[str(i['description']),str(timedelta(seconds = int(((time.time()*1000)-int(i['start_timestamp_ms']))/1000)))] for i in running]
    now = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
    if credential_name != None:
      print(credential_name)
    print(len(ready),'tasks ready',now)
    print(len(running),'tasks running',now)
    print('Running names:')
    for rn in running_names:print(rn)
    print()
    print()
    time.sleep(5)
    task_count = len(ready) +len(running) 
#Get tasks in an easy-to-use format
def getTasks(id_list = None):
    tasks = ee.data.getTaskList()
    if id_list != None:
      tasks = [i for i in tasks if i['description'] in id_list]
    out = {}
    out['ready'] = [i['description'] for i in tasks if i['state'] == 'READY']
    out['running'] = [i['description'] for i in tasks if i['state'] == 'RUNNING']
    out['failed'] = [i['description'] for i in tasks if i['state'] == 'FAILED']
    out['completed'] = [i['description'] for i in tasks if i['state'] == 'COMPLETED']
    return out
# Standard task tracker - prints number of ready and running tasks each 10 seconds
def failedTasks():
    tasks = ee.data.getTaskList()
    failed = [i for i in tasks if i['state'] == 'FAILED']
    failed_names = [[str(i['description']),str(timedelta(seconds = int(((time.time()*1000)-int(i['start_timestamp_ms']))/1000)))] for i in failed]
    now = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
    print('Failed names:')
    for rn in failed: print(rn)



#------------------------------------------------------------------------------
#                  Cancel Tasks
#------------------------------------------------------------------------------
# Cancels all running tasks
def batchCancel():
    tasks = ee.data.getTaskList()
    cancelledTasks = []
    for ind, i in enumerate(tasks):
        if i['state'] == 'READY' or i['state'] == 'RUNNING':
            cancelledTasks.append(ind)
            ee.data.cancelTask(i['id'])
            #ee.data.cancelOperation(ee._cloud_api_utils.convert_task_id_to_operation_name(i['id'])) # this was the workaround for a bug with cancelTask in earthengine-api v0.1.225
    tasks2 = ee.data.getTaskList()
    for ind in cancelledTasks:
        print(tasks2[ind]['state']+': '+tasks2[ind]['description'])

# Cancels all running tasks with identifier in their name
def cancelByName(nameIdentifier):
    cancelList = nameTaskList(nameIdentifier)
    if cancelList:
        for task in cancelList:
            print('Cancelling '+task['description'])
            ee.data.cancelTask(task['id'])
    else:
        print('No Tasks to Cancel')
#------------------------------------------------------------------------------
#                 Task Tracking Using Time Interval
#------------------------------------------------------------------------------    
# Get list of tasks started within a specified time interval
# Starttime and endtime must be python datetimes, in UTC, e.g. datetime.datetime.utcnow()
def timeTaskList(starttime, endtime):
    tasks = ee.data.getTaskList()
    epoch = datetime.utcfromtimestamp(0)
    starttime = (starttime - epoch).total_seconds() * 1000.
    endtime = (endtime - epoch).total_seconds() * 1000.
    thisList = [i for i in tasks if (i['creation_timestamp_ms'] >= starttime and i['creation_timestamp_ms'] <= endtime)]
    return thisList

# Using a start and end time, track the jobs that were started within that time interval
# Print status updates every minute
# Only move on when no 'ready' or 'running' jobs remain, i.e. all have completed or failed
# Then print status of all tasks 
# Starttime and endtime must be python datetimes, in UTC, e.g. datetime.datetime.utcnow()
# check_interval = seconds between status checks
def jobCompletionTracker(starttime, endtime, check_interval):
    timeStart = datetime.utcnow()
    thisTasklist = timeTaskList(starttime,endtime)
    for i in thisTasklist:
        print(i['description'], i['state'])
    currentJobs = 1
    
    while currentJobs > 0:   
        time.sleep(check_interval) 
        thisTasklist = timeTaskList(starttime,endtime)
        ready = [i for i in thisTasklist if (i['state'] == 'READY')]
        running = [i for i in thisTasklist if (i['state'] == 'RUNNING')]
        completed = [i for i in thisTasklist if (i['state'] == 'COMPLETED')]
        failed = [i for i in thisTasklist if (i['state'] == 'FAILED')]
        cancelled = [i for i in thisTasklist if (i['state'] == 'CANCELLED')]
        currentJobs = len(ready) + len(running)
        
        print(' ')
        print(datetime.now(), ':')
        print(len(ready),' tasks ready')
        #print(len(running), ' tasks running')
        running_names = [[str(i['description']),str(timedelta(seconds = int(((time.time()*1000)-int(i['start_timestamp_ms']))/1000)))] for i in running]
        for rn in running_names:
            print(rn)

    timeEnd = datetime.utcnow()
    print('Process Time:')
    print(timeEnd - timeStart)          
    return [ready, running, completed, failed, cancelled, thisTasklist]

#------------------------------------------------------------------------------
#                 Task Tracking Using Task Name / Description
#------------------------------------------------------------------------------    
# Get list of tasks with specified name/description prefix (e.g. "description" passed to GEE export function)
def nameTaskList(nameIdentifier):
    tasks = ee.data.getTaskList()
    thisList = [i for i in tasks if re.findall(nameIdentifier, i['description']) and (i['state'] == 'READY' or i['state'] == 'RUNNING')]
    return thisList


import ee,os,json,re
ee.Initialize()
###################################################################################################
# This script recursively migrates Google Earth Engine assets from one repository to a new one, sets permissions, and deletes old assets if needed.
# **Important** for this to work you need the following:
#       - The credential you're using must have editor permissions in both the source and destination repositories
#                               OR
#       - The credential you're using must have editor permissions in the destination repository, and you are willing to
#           change the permissions of all the original assets to "Anyone Can Read" (to do this, set changePermissions = True below)
#       - The geeViz package installed in your python instance. (https://github.com/gee-community/geeViz)

#Written by:
#Ian Housman ian.housman@usda.gov ian.housman@gmail.com 
#Leah Campbell leah.campbell@usda.gov leahs.campbell@gmail.com
#RedCastle Resources Inc.
#MIT License
###################################################################################################
# Repository you are moving from:
sourceRoot = 'users/iwhousman/test'
# sourceRoot = 'projects/USFS/LCMS-NFS/R1/FNF'
# sourceRoot = 'projects/lcms-292214/assets/AK-Ancillary-Data'

# Repository you are moving to:
# It is assumed this folder already exists and user account has write access to it
# destinationRoot = 'projects/lcms-292214/assets/migrationTest/FNF'
# destinationRoot = 'users/usfs-ihousman/migrationTest'
# If the credential you're using does not have editor permissions in the source repository
#Must include 'user:' prefix if it is a individual's Email or 'group:' if it is a Google Group
changeSourcePermissions = True
sourceReaders = []
sourceWriters = ['user:ian.housman@usda.gov']
source_all_users_can_read = False

changeDestinationPermissions = True
destinationReaders = []
destinationWriters = []
destination_all_users_can_read = False
###################################################################################################
#Function to get all folders, imageCollections, images, and tables under a given folder or imageCollection level
def getTree(fromRoot,toRoot,treeList = []):
    pathPrefix = 'projects/earthengine-legacy/assets/'

    #Clean up the given paths
    if(fromRoot[-1] != '/'):fromRoot += '/'
    if(toRoot[-1] != '/'):toRoot += '/'

    #Handle inconsistencies with the earthengine-legacy prefix
    if re.match("^projects/[^/]+/assets/.*$",fromRoot) == None:
        fromRoot = pathPrefix + fromRoot
    if re.match("^projects/[^/]+/assets/.*$",toRoot) == None:
        toRoot = pathPrefix + toRoot
    # print(fromRoot,toRoot)
    #List assets 
    try:
        assets = ee.data.listAssets({'parent':fromRoot})['assets']
    except Exception as e1:
        print(e1)
        # try:
        #     assets = ee.data.listAssets({'parent':fromRoot})['assets']
        # except Exception as e2:
        #     print(e2)

    #Reursively walk down the tree
    nextLevels = []
    for asset in assets:
        fromID = asset['name']
        fromType = asset['type']
        # fromID = fromID.replace('projects/earthengine-legacy/assets/','')
        toID = fromID.replace(fromRoot,toRoot)
        
        if fromType in ['FOLDER','IMAGE_COLLECTION']:
            nextLevels.append([fromID,toID])
        treeList.append([fromType,fromID,toID])  
    
    
    for i1,i2 in nextLevels:
        getTree(i1,i2,treeList)
    return treeList
###################################################################################################
#Function for setting permissions for all files under a specified root level
#Either a list of assets a root to start from can be provided
def batchChangePermissions(assetList = None,root = None,readers = [],writers = [], all_users_can_read = False):
    if assetList == None:
        assetList = [i[1] for i in getTree(root,root)]

    for assetID in assetList:
        print('Changing permissions for: {}'.format(assetID))
        try:
            ee.data.setAssetAcl(assetID, json.dumps({u'writers': writers, u'all_users_can_read': all_users_can_read, u'readers': readers}))
        except Exception as e:
            print(e)
###################################################################################################
#Function to copy all folders, imageCollections, images, and tables under a given folder or imageCollection level
#Permissions can also be set here 
def copyAssetTree(fromRoot,toRoot,changePermissions = False,readers = [],writers = [],all_users_can_read = False):
    treeList = getTree(fromRoot,toRoot)
    
    #Iterate across all assets and copy and create when appropriate
    for fromType,fromID,toID in treeList:
        if fromType in ['FOLDER','IMAGE_COLLECTION']:
            try:
                print('Creating {}: {}'.format(fromType,toID))
                ee.data.createAsset({'type':'Image_Collection', 'name': toID})
            except Exception as e:
                print(e)
        else:
            try:
                print('Copying {}: {}'.format(fromType,toID))
                ee.data.copyAsset(fromID,toID,False)

            except Exception as e:
                print(e)
        print()
        print()

    if changePermissions:
        batchChangePermissions(assetList = [i[2] for i in treeList],root = None,readers = [],writers = [], all_users_can_read = False)
###################################################################################################
#Function to delete all folders, imageCollections, images, and tables under a given folder or imageCollection level
def deleteAssetTree(root):
    answer = input('Are you sure you want to delete all assets under {}? (y = yes, n = no) '.format(root))
    print(answer)
    if answer.lower() == 'y':
        answer = input('You answered yes. Just double checking. Are you sure you want to delete all assets under {}? (y = yes, n = no) '.format(root))
        if answer.lower() == 'y':
            treeList = getTree(root,root)
            treeList.reverse()
            for fromType, ID1,ID2 in treeList:
                print('Deleting {}'.format(ID1))
                try:
                    ee.data.deleteAsset(ID1)
                except Exception as e:
                    print(e)
###################################################################################################
#Step 1: Make sure source has account added as reader (may need to use different credentials for this)
batchChangePermissions(None, sourceRoot,sourceReaders,sourceWriters, source_all_users_can_read)
###################################################################################################
#Step 2: Use this function to copy assets
# copyAssetTree(sourceRoot,destinationRoot,changeDestinationPermissions,destinationReaders,destinationWriters,destination_all_users_can_read)
###################################################################################################
#!!!!!!! DANGER !!!!!!!!!
#!!!!!!! DANGER !!!!!!!!!
#!!!!!!! DANGER !!!!!!!!!
#!!!!!!! DANGER !!!!!!!!!
#Optional Step 3: Once all assets are copied and inspected, you can use this method to delete all files under the root level
#This method is final and there is no way to undo this
# deleteAssetTree(destinationRoot)


"""
   Copyright 2021 Ian Housman

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

#Example of how to get Climate data using the getImagesLib and view outputs using the Python visualization tools
#Acquires DAYMET V4 data and then adds them to the viewer
####################################################################################################
import os,sys
sys.path.append(os.getcwd())

#Module imports
import geeViz.getImagesLib as getImagesLib
ee = getImagesLib.ee
Map = getImagesLib.Map
Map.clearMap()
####################################################################################################
#Define user parameters:

# Specify study area: Study area
# Can be a featureCollection, feature, or geometry
studyArea = getImagesLib.testAreas['CA']

# Update the startJulian and endJulian variables to indicate your seasonal 
# constraints. This supports wrapping for tropics and southern hemisphere.
# If using wrapping and the majority of the days occur in the second year, the system:time_start will default 
# to June 1 of that year.Otherwise, all system:time_starts will default to June 1 of the given year
# startJulian: Starting Julian date 
# endJulian: Ending Julian date
startJulian = 274
endJulian = 273

# Specify start and end years for all analyses
# More than a 3 year span should be provided for time series methods to work 
# well. If providing pre-computed stats for cloudScore and TDOM, this does not 
# matter
startYear = 2017
endYear = 2019

# Specify an annual buffer to include imagery from the same season 
# timeframe from the prior and following year. timeBuffer = 1 will result 
# in a 3 year moving window. If you want single-year composites, set to 0
timebuffer =0

# Specify the weights to be used for the moving window created by timeBuffer
# For example- if timeBuffer is 1, that is a 3 year moving window
# If the center year is 2000, then the years are 1999,2000, and 2001
# In order to overweight the center year, you could specify the weights as
# [1,5,1] which would duplicate the center year 5 times and increase its weight for
# the compositing method. If timeBuffer = 0, set to [1]
weights = [1]

# Choose reducer to use for summarizing each year
# Common options are ee.Reducer.mean() and ee.Reducer.median()
compositingReducer = ee.Reducer.mean()

# Choose collection to use
# Supports:
# NASA/ORNL/DAYMET_V3
# NASA/ORNL/DAYMET_V4
# UCSB-CHG/CHIRPS/DAILY (precipitation only)
collectionName = 'NASA/ORNL/DAYMET_V4'



# Export params
# Whether to export climate composites
exportComposites = True


# Provide location composites will be exported to
# This should be an asset folder, or more ideally, an asset imageCollection
exportPathRoot = 'users/iwhousman/test/climate-test'

# Specify which bands to export
# If not sure or want all bands, just set to None
exportBands = ['prcp.*','tmax.*','tmin.*']

# CRS- must be provided.  
# Common crs codes: Web mercator is EPSG:4326, USGS Albers is EPSG:5070, 
# WGS84 UTM N hemisphere is EPSG:326+ zone number (zone 12 N would be EPSG:32612) and S hemisphere is EPSG:327+ zone number
crs = 'EPSG:5070'

# Specify transform if scale is null and snapping to known grid is needed
transform = [1000,0,-2361915.0,0,-1000,3177735.0]

# Specify scale if transform is null
scale = None
####################################################################################################
#End user parameters
####################################################################################################
####################################################################################################
####################################################################################################
#Start function calls
####################################################################################################
####################################################################################################
#Call on master wrapper function to get climate composites
climateComposites = getImagesLib.getClimateWrapper(collectionName,studyArea,startYear,endYear,startJulian,endJulian,timebuffer,weights,compositingReducer,exportComposites,exportPathRoot,crs,transform,scale,exportBands)

Map.addTimeLapse(climateComposites.select(exportBands),{},'Climate Composite Time Lapse')

####################################################################################################
#Load the study region
Map.addLayer(studyArea, {'strokeColor': '0000FF'}, "Study Area", True)
Map.centerObject(studyArea)
####################################################################################################
####################################################################################################
Map.view()


'''
Based on https://github.com/google/earthengine-community/blob/master/datasets/scripts/LCMS_Visualization.js
 * Copyright 2021 The Google Earth Engine Community Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * 
 * Example script for visualizing LCMS change summaries, land cover, and land use.
 * 
 * A more in-depth visualization of LCMS products is available at:
 * https://apps.fs.usda.gov/lcms-viewer/
 * 
 * Contact sm.fs.lcms@usda.gov with any questions or specific data requests.
''' 
####################################################################################################
#Import modules
import os,sys
sys.path.append(os.getcwd())

import  geeViz.geeView as geeView
ee = geeView.ee
Map = geeView.Map

Map.clearMap()
####################################################################################################
#############################################################################
### Define visualization parameters ###
#############################################################################
startYear = 1985;
endYear = 2020;
lossYearPalette = ['ffffe5', 'fff7bc', 'fee391', 'fec44f', 'fe9929',
                   'ec7014', 'cc4c02']
gainYearPalette = ['c5ee93', '00a398']
durationPalette = ['BD1600', 'E2F400','0C2780']

lossYearViz = {'min': startYear, 'max': endYear, 'palette': lossYearPalette};
gainYearViz = {'min': startYear, 'max': endYear, 'palette': gainYearPalette};
durationViz = {'min': 1, 'max': 5, 'palette': durationPalette};
#############################################################################
### Define functions ###
#############################################################################

#Convert given code to year that number was present in the image.

def getMostRecentChange(c, code):
	def wrapper(img):
		yr = ee.Date(img.get('system:time_start')).get('year')
		return ee.Image(yr).int16().rename(['year']).updateMask(img.eq(code)).copyProperties(img,['system:time_start'])
	return c.map(wrapper)
  

#############################################################################
### Bring in LCMS annual outputs ###
#############################################################################

lcms = ee.ImageCollection('USFS/GTAC/LCMS/v2020-5');
bandNames =  lcms.first().bandNames().getInfo()
print('Available study areas:', lcms.aggregate_histogram('study_area').keys().getInfo());
print('Available LCMS products',bandNames);
print('Learn more about visualization of LCMS products here', 'https://apps.fs.usda.gov/lcms-viewer/')

#Filter out study area
lcms = lcms.filter(ee.Filter.eq('study_area','CONUS'))

#Set up time periods to compare land cover and land use
earlySpan = [startYear, startYear+4]
lateSpan = [endYear-4, endYear]

#############################################################################
### Add full raw model outputs ###
#############################################################################

#Separate products
raw_change = lcms.select(['Change_Raw_.*'])
raw_land_cover = lcms.select(['Land_Cover_Raw_.*'])
raw_land_use = lcms.select(['Land_Use_Raw_.*'])

#Shorten names
raw_change_bns = [bn for bn in bandNames if bn.find('Change_Raw_') > -1]
raw_land_cover_bns = [bn for bn in bandNames if bn.find('Land_Cover_Raw_') > -1]
raw_land_use_bns = [bn for bn in bandNames if bn.find('Land_Use_Raw_') > -1]

raw_change_bns_short = [i.split('_Probability_')[-1] for i in raw_change_bns]
raw_land_cover_bns_short = [i.split('_Probability_')[-1] for i in raw_land_cover_bns]
raw_land_use_bns_short = [i.split('_Probability_')[-1] for i in raw_land_use_bns]

raw_change = raw_change.select(raw_change_bns,raw_change_bns_short)
raw_land_cover = raw_land_cover.select(raw_land_cover_bns,raw_land_cover_bns_short)
raw_land_use = raw_land_use.select(raw_land_use_bns,raw_land_use_bns_short)


#Add to map
Map.addLayer(raw_change,{'min':0,'max':30,'opacity':0,'addToLegend':False},'Raw LCMS Change Model Probability',True) 
Map.addLayer(raw_land_cover,{'min':0,'max':30,'opacity':0,'addToLegend':False},'Raw LCMS Land Cover Model Probability',True) 
Map.addLayer(raw_land_use,{'min':0,'max':30,'opacity':0,'addToLegend':False},'Raw LCMS Land Use Model Probability',True) 

#############################################################################
### Visualize Land Use change ###
#############################################################################
#Copy properties from original image after reducing collection to retain visualization properties
lu = lcms.select(['Land_Use'])
earlyLu = lu.filter(ee.Filter.calendarRange(earlySpan[0], earlySpan[1], 'year')).mode().copyProperties(lcms.first())
lateLu = lu.filter(ee.Filter.calendarRange(lateSpan[0], lateSpan[1], 'year')).mode().copyProperties(lcms.first())
Map.addLayer(earlyLu, {'autoViz':True}, 'Early Land Use Mode ({}-{})'.format(earlySpan[0],earlySpan[1]), False)
Map.addLayer(lateLu, {'autoViz':True}, 'Recent Land Use Mode ({}-{})'.format(lateSpan[0],lateSpan[1]), False);



#############################################################################
### Visualize Land Cover change ###
#############################################################################
#Copy properties from original image after reducing collection to retain visualization properties
lc = lcms.select(['Land_Cover'])
earlyLc = lc.filter(ee.Filter.calendarRange(earlySpan[0], earlySpan[1], 'year')).mode().copyProperties(lcms.first())
lateLc = lc.filter(ee.Filter.calendarRange(lateSpan[0], lateSpan[1], 'year')).mode().copyProperties(lcms.first())
Map.addLayer(earlyLc, {'autoViz':True}, 'Early Land Cover Mode ({}-{})'.format(earlySpan[0],earlySpan[1]), False);
Map.addLayer(lateLc, {'autoViz':True}, 'Recent Land Cover Mode ({}-{})'.format(lateSpan[0],lateSpan[1]), False);



#############################################################################
### Visualize Change products ###
#############################################################################

#Select the change band. Land_Cover and Land_Use are also available.
change = lcms.select(['Change'])

#Convert to year collection for a given code.
slowLossYears = getMostRecentChange(change, 2)
fastLossYears = getMostRecentChange(change, 3)
gainYears = getMostRecentChange(change, 4)

#Find the most recent year.
mostRecentSlowLossYear = slowLossYears.max()
mostRecentFastLossYear = fastLossYears.max()
mostRecentGainYear = gainYears.max()

#Find the duration.
slowLossDuration = slowLossYears.count()
fastLossDuration = fastLossYears.count()
gainDuration = gainYears.count()

#Add year summaries to the map.
Map.addLayer(mostRecentSlowLossYear, lossYearViz, 'Most Recent Slow Loss Year', True)
Map.addLayer(mostRecentFastLossYear, lossYearViz, 'Most Recent Fast Loss Year', True)
Map.addLayer(mostRecentGainYear, gainYearViz, 'Most Recent Gain Year', True)

#Add durations to the map.
Map.addLayer(slowLossDuration,durationViz, 'Slow Loss Duration', False);
Map.addLayer(fastLossDuration,durationViz, 'Fast Loss Duration', False);
Map.addLayer(gainDuration,durationViz, 'Gain Duration', False);

justChange = change.map(lambda img:img.updateMask(ee.Image(img.gt(1).And(img.lt(5))).copyProperties(lcms.first())))
Map.addTimeLapse(justChange,{'autoViz':True},'LCMS Change Time Lapse',False)


##############################################################################
#### Map setup ###
##############################################################################
Map.centerObject(lcms.filter(ee.Filter.eq('study_area', 'CONUS')).first().geometry())
Map.turnOnInspector()
Map.view()